//
//
//  Simple Texture Buffer Object
//
//    This sample shows how to create and use texture buffer objects as defined in
//  GL_EXT_texture_buffer_object. This sample uses the texture buffer objects as sources
//  for vertex data, allowing independent indexing of vertex and normal data. Regular VBO
//  rendering is provided for comparison, and it is expected to be faster. More typically
//  texture buffer objects serve as a large source of constant data for operations like
//  skinning and instancing.
//
// Author: Evan Hart
// Email: sdkfeedback@nvidia.com
//
// Copyright (c) NVIDIA Corporation. All rights reserved.
////////////////////////////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <map>

#include <GL/glew.h>
#include <GL/glut.h>

#include <nvMath.h>
#include <nvModel.h>
#include <nvGlutManipulators.h>

#define NV_REPORT_COMPILE_ERRORS

#include <nvShaderUtils.h>
#include <nvSDKPath.h>
#include <nvGlutWidgets.h>

using std::map;

////////////////////////////////////////////////////////////////////////////////
//
// Globals
//
////////////////////////////////////////////////////////////////////////////////

int width = 512;
int height = 512;

//data paths
#define MODEL_PATH "media/models/"
#define SHADER_PATH "src/simple_texture_buffer_object/shaders/"

//model and the default model file
nv::Model *model;
const char modelFile[] = MODEL_PATH "venusm.obj";

nv::SDKPath sdkPath;

float diameter = 0.0f; //diameter of the model
nv::vec3f center(0.0f, 0.0f, 0.0f); //center of the model

//model controller
nv::GlutExamine manipulator;

enum UIOption {
    OPTION_DISPLAY_WIREFRAME,
    OPTION_USE_TEXBO,
    OPTION_ANIMATE,
    OPTION_COUNT
};
bool options[OPTION_COUNT];
map<char, UIOption> optionKeyMap;

nv::GlutUIContext ui;

//vertex buffer objects and texture buffers used for rendering
GLuint textures[2];
GLuint textureBuffers[2];
GLuint vbo;
GLuint indexBuffer[2];

//shaders used to display the object
GLuint vShader, vShaderTexBO;
GLuint fShader;
GLuint program, programTexBO;

//shader files
const char texBOVShaderFile[] = SHADER_PATH "texBO_vertex.glsl";
const char regularVShaderFile[] = SHADER_PATH "regular_vertex.glsl";
const char fShaderFile[] = SHADER_PATH "regular_fragment.glsl";


////////////////////////////////////////////////////////////////////////////////
//
// Functions
//
////////////////////////////////////////////////////////////////////////////////

//
// keys
//  Callback function for keyboard input.
//////////////////////////////////////////////////////////////////////
void keys( unsigned char c, int x, int y) {
    ui.keyboard(c, x, y);

    c = (unsigned char)tolower(c);

    //toggle the option, if we recognize it
    if ( optionKeyMap.find(c) != optionKeyMap.end())
        options[optionKeyMap[c]] = ! options[optionKeyMap[c]];

    switch (c) {

        case '\033': //escape key
        case 'q':
            exit(0);
    }

    glutPostRedisplay();
}

//
// special
//  Callback function for special keyboard input.
//////////////////////////////////////////////////////////////////////
void special( int c, int x, int y) {
    ui.keyboard(c, x, y);
}

//
//
//////////////////////////////////////////////////////////////////////
inline void updateButtonState( const nv::ButtonState &bs, nv::GlutManipulator &manip, int button) {
    int modMask = 0;

    if (bs.state & nv::ButtonFlags_Alt) modMask |= GLUT_ACTIVE_ALT;
    if (bs.state & nv::ButtonFlags_Shift) modMask |= GLUT_ACTIVE_SHIFT;
    if (bs.state & nv::ButtonFlags_Ctrl) modMask |= GLUT_ACTIVE_CTRL;

    if (bs.state & nv::ButtonFlags_End)
        manip.mouse( button, GLUT_UP, modMask, bs.cursor.x, height - bs.cursor.y);
    if (bs.state & nv::ButtonFlags_Begin)
        manip.mouse( button, GLUT_DOWN, modMask, bs.cursor.x, height - bs.cursor.y);
}

//
//
//////////////////////////////////////////////////////////////////////
void doUI()
{
	nv::Rect none(0,0), none1(0,0);
	int radioSelection = (int)(!options[OPTION_USE_TEXBO]);

    glDisable( GL_CULL_FACE);

	ui.begin();

    ui.beginGroup();
	    ui.doLabel(none, "Simple Texture Buffer Object");
        ui.doRadioButton( 0, none1, "Texture buffers", &radioSelection);
        ui.doRadioButton( 1, none, "Vertex buffers", &radioSelection);
        ui.doCheckButton(none, "Wireframe", &options[OPTION_DISPLAY_WIREFRAME]);
    ui.endGroup();

    options[OPTION_USE_TEXBO] = radioSelection ? false : true;

    // Pass non-ui mouse events to the manipulator
    if (!ui.isOnFocus()) {
        const nv::ButtonState &lbState = ui.getMouseState( 0);
        const nv::ButtonState &mbState = ui.getMouseState( 1);
        const nv::ButtonState &rbState =  ui.getMouseState( 2);
        
        manipulator.motion( ui.getCursorX(), height - ui.getCursorY());

        updateButtonState( lbState, manipulator, GLUT_LEFT_BUTTON);
        updateButtonState( mbState, manipulator, GLUT_MIDDLE_BUTTON);
        updateButtonState( rbState, manipulator, GLUT_RIGHT_BUTTON);
    }

	ui.end();
}

//
// draw
//  Callback function used for all rendering
//////////////////////////////////////////////////////////////////////
void draw() {
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // setup a simple projection matrix based on the size of the model
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective( 40.0, (double)width/height, diameter*0.05, diameter*20.5);

    // setup the model matrix with the current transform from the controlled after
    // centering the model
    glMatrixMode( GL_MODELVIEW);
    glLoadIdentity();
    manipulator.applyTransform();
    glTranslatef( -center[0], -center[1], -center[2]);

    // setup for either solid or wireframe drawing
    if (options[OPTION_DISPLAY_WIREFRAME]) {
        glPolygonMode( GL_FRONT_AND_BACK, GL_LINE);
        glDisable(GL_CULL_FACE);
    }
    else {
        glPolygonMode( GL_FRONT_AND_BACK, GL_FILL);
        glEnable(GL_CULL_FACE);
    }

    if (options[OPTION_USE_TEXBO]) {

        // bind the texture buffer objects containing the normals and the vertices
        glActiveTexture( GL_TEXTURE0);
        glBindTexture( GL_TEXTURE_BUFFER_EXT, textures[0]);
        glActiveTexture( GL_TEXTURE1);
        glBindTexture( GL_TEXTURE_BUFFER_EXT, textures[1]);

        // bind the merged set of indices as the vertex buffer
        glBindBuffer( GL_ARRAY_BUFFER, indexBuffer[0]);
        glVertexPointer( 2, GL_FLOAT, 0, 0);
        glEnableClientState( GL_VERTEX_ARRAY);

        glUseProgram( programTexBO);

        // since the indexing happens in the shader, the indices are sent directly
        // as vertices
        glDrawArrays( GL_TRIANGLES, 0, model->getIndexCount());

        glDisableClientState( GL_VERTEX_ARRAY);
        glUseProgram( 0);

        glBindBuffer( GL_ARRAY_BUFFER, 0);
    }
    else {

        // bind an configure the vertex buffer for rendering
        glBindBuffer( GL_ARRAY_BUFFER, vbo);
        glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, indexBuffer[1]);

        glVertexPointer( model->getPositionSize(), GL_FLOAT, model->getCompiledVertexSize() * sizeof(float), 0);
        glNormalPointer( GL_FLOAT, model->getCompiledVertexSize() * sizeof(float), (void*)(model->getCompiledNormalOffset() * sizeof(float)));
        
        glEnableClientState( GL_VERTEX_ARRAY);
        glEnableClientState( GL_NORMAL_ARRAY);

        glUseProgram( program);

        glDrawElements( GL_TRIANGLES, model->getCompiledIndexCount(), GL_UNSIGNED_INT, 0);

        glUseProgram( 0);

        glDisableClientState( GL_VERTEX_ARRAY);
        glDisableClientState( GL_NORMAL_ARRAY);

        glBindBuffer( GL_ARRAY_BUFFER, 0);
        glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, 0);
    }

    doUI();

    glutSwapBuffers();
}

//
// idle
//  Callback used to keep the screen redrawing
//////////////////////////////////////////////////////////////////////
void idle() {
    if (options[OPTION_ANIMATE]) {
        //increment the rotation
        manipulator.idle();
    }

    glutPostRedisplay();
}

//
// reshape
//  Callack for any change in window size
//////////////////////////////////////////////////////////////////////
void reshape( int w, int h) {

    ui.reshape(w, h);

    width = w;
    height = h;
    glViewport( 0, 0, width, height);

    //pass the change along to the controller
    manipulator.reshape( w, h);
}



//
//
//////////////////////////////////////////////////////////////////////
void mouse(int button, int state, int x, int y) {
    ui.mouse(button, state, glutGetModifiers(), x, y);
}

//
//
//////////////////////////////////////////////////////////////////////
void motion(int x, int y) {
    ui.mouseMotion(x, y);
}

//
//
//////////////////////////////////////////////////////////////////////
void passiveMotion(int x, int y) {
	ui.mouseMotion(x, y);
}

//
//
//////////////////////////////////////////////////////////////////////
void init() {
    std::string resolved_path;
    model = new nv::Model;

    if (sdkPath.getFilePath( modelFile, resolved_path)) {
        if (!model->loadModelFromFile( resolved_path.c_str())) {
            delete model;
            model = NULL;
            printf( "Error: Failed to load model: %s\n", modelFile);
            exit(-1);
        }
    }
    else {
        printf( "Error: Failed to find model file: %s\n", modelFile);
        exit(-1);
    }

    glGenTextures( 2, textures);
    glGenBuffers( 2, textureBuffers);
    glGenBuffers( 2, indexBuffer);
    glGenBuffers( 1, &vbo);

    //compute the model dimensions
    nv::vec3f minPos, maxPos;

    model->computeBoundingBox( minPos, maxPos);
    center = (minPos + maxPos) * 0.5f;
    diameter = nv::length( maxPos - minPos);

    manipulator.setDollyPosition( -1.5f * diameter);

    //create and fill the texture buffers

    //vertex texture buffer first
    glBindBuffer( GL_TEXTURE_BUFFER_EXT, textureBuffers[0]);

    // must pad positions to size 4 for texbo efficiency
    if ( model->getPositionSize() == 3) {
        
        float *tempPos = new float[model->getPositionCount() * 4];
        const float *src = model->getPositions();
        float *dst = tempPos;
        for (int ii = 0; ii<model->getPositionCount(); ii++) {
            *dst++ = *src++;
            *dst++ = *src++;
            *dst++ = *src++;
            *dst++ = 1.0f;
        }
        glBufferData( GL_TEXTURE_BUFFER_EXT, 4 * model->getPositionCount() * sizeof(float), tempPos, GL_STATIC_DRAW);
        delete []tempPos;
    }
    else {
        glBufferData( GL_TEXTURE_BUFFER_EXT, model->getPositionSize() * model->getPositionCount() * sizeof(float), model->getPositions(), GL_STATIC_DRAW);
    }
    glBindTexture( GL_TEXTURE_BUFFER_EXT, textures[0]);
    glTexBufferEXT( GL_TEXTURE_BUFFER_EXT, GL_RGBA32F_ARB, textureBuffers[0]);

    //normal texture buffer first
    glBindBuffer( GL_TEXTURE_BUFFER_EXT, textureBuffers[1]);
    // must pad normals to size 4 for texbo efficiency
    if ( model->getNormalSize() == 3) {
        float *tempNorm = new float[model->getNormalCount() * 4];
        const float *src = model->getNormals();
        float *dst = tempNorm;
        for (int ii = 0; ii<model->getNormalCount(); ii++) {
            *dst++ = *src++;
            *dst++ = *src++;
            *dst++ = *src++;
            *dst++ = 1.0f;
        }
        glBufferData( GL_TEXTURE_BUFFER_EXT, 4 * model->getNormalCount() * sizeof(float), tempNorm, GL_STATIC_DRAW);
        delete []tempNorm;
    }
    else {
        glBufferData( GL_TEXTURE_BUFFER_EXT, model->getNormalSize() * model->getNormalCount() * sizeof(float), model->getNormals(), GL_STATIC_DRAW);
    }
    glBindTexture( GL_TEXTURE_BUFFER_EXT, textures[1]);
    glTexBufferEXT( GL_TEXTURE_BUFFER_EXT, GL_RGBA32F_ARB, textureBuffers[1]);

    //create the index list
    float *indices = new float[ model->getIndexCount() * 2];

    const GLuint *vtxIndex = model->getPositionIndices();
    const GLuint *nrmIndex = model->getNormalIndices();
    float *wlk = indices;

    for (int ii=0; ii<model->getIndexCount(); ii++) {
        *wlk++ = *vtxIndex++;
        *wlk++ = *nrmIndex++;
    }

    glBindBuffer( GL_ARRAY_BUFFER, indexBuffer[0]);
    glBufferData( GL_ARRAY_BUFFER, sizeof(GLuint) * model->getIndexCount() * 2, indices, GL_STATIC_DRAW);

    delete []indices;

    //now, compute a compiled model or reference
    model->compileModel();

    // load the vertex data for the compiled model
    glBindBuffer( GL_ARRAY_BUFFER, vbo);
    glBufferData( GL_ARRAY_BUFFER, model->getCompiledVertexSize() * model->getCompiledVertexCount() * sizeof(float), model->getCompiledVertices(), GL_STATIC_DRAW);

    //load the index data for the compiled model
    glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, indexBuffer[1]);
    glBufferData( GL_ELEMENT_ARRAY_BUFFER, model->getCompiledIndexCount() * sizeof(GLuint), model->getCompiledIndices(), GL_STATIC_DRAW);

    glBindBuffer( GL_ARRAY_BUFFER, 0);
    glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, 0);
    glBindBuffer( GL_TEXTURE_BUFFER_EXT, 0);


    //load shaders
    if (sdkPath.getFilePath( regularVShaderFile, resolved_path)) {
        vShader = nv::CompileGLSLShaderFromFile( GL_VERTEX_SHADER, resolved_path.c_str());
        if ( vShader == 0) printf( "Vertex shader failed to compile\n");
    }
    else {
        printf( "Failed to locate vertex shader\n");
    }

    if (sdkPath.getFilePath( texBOVShaderFile, resolved_path)) {
        vShaderTexBO = nv::CompileGLSLShaderFromFile( GL_VERTEX_SHADER, resolved_path.c_str());
        if ( vShaderTexBO == 0) printf( "TexBO vertex shader failed to compile\n");
    }
    else {
        printf( "Failed to locate TexBO vertex shader\n");
    }

    if (sdkPath.getFilePath( fShaderFile, resolved_path)) {
        fShader = nv::CompileGLSLShaderFromFile( GL_FRAGMENT_SHADER, resolved_path.c_str());
        if ( fShader == 0) printf( "Fragment shader failed to compile\n");
    }
    else {
        printf( "Failed to locate fragment shader\n");
    }

    program = glCreateProgram();
    glAttachShader( program, vShader);
    glAttachShader( program, fShader);

    glLinkProgram( program);

    // check if program linked
    GLint success = 0;
    glGetProgramiv( program, GL_LINK_STATUS, &success);

    if (!success) {
        char temp[256];
        glGetProgramInfoLog( program, 256, NULL, temp);
        printf( "Failed to link program:\n%s\n");
        glDeleteProgram(program);
        program = 0;
    }

    programTexBO = glCreateProgram();
    glAttachShader( programTexBO, vShaderTexBO);
    glAttachShader( programTexBO, fShader);

    glLinkProgram( programTexBO);

    // check if program linked
    glGetProgramiv( programTexBO, GL_LINK_STATUS, &success);

    if (!success) {
        char temp[256];
        glGetProgramInfoLog( programTexBO, 256, NULL, temp);
        printf( "Failed to link program:\n%s\n");
        glDeleteProgram(programTexBO);
        programTexBO = 0;
    }
    else {
        glUseProgram( programTexBO);

        // Set the samplers
        glUniform1i( glGetUniformLocation( programTexBO, "vertexSampler"), 0);
        glUniform1i( glGetUniformLocation( programTexBO, "normalSampler"), 1);

        glUseProgram( 0);
    }


    glEnable(GL_DEPTH_TEST);
    glDepthFunc( GL_LEQUAL);

}

//
//
//////////////////////////////////////////////////////////////////////
int main( int argc, char** argv) {
    glutInit( &argc, argv);
    glutInitDisplayMode( GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
    glutInitWindowSize( width, height);

    glutCreateWindow( "Simple Texture Buffer Object");

    glewInit();

    if (! glewIsSupported( "GL_VERSION_2_0 GL_EXT_gpu_shader4 GL_EXT_texture_buffer_object")) {
        printf( "Required extensions not supported.\n");
        printf( "This sample requires:\n");
        printf( "  OpenGL version 2.0 or later\n");
        printf( "  GL_EXT_gpu_shader4\n");
        printf( "  GL_EXT_texture_buffer_object\n");
        return 1;
    }

    //configure the options
    optionKeyMap['w'] = OPTION_DISPLAY_WIREFRAME;
    options[OPTION_DISPLAY_WIREFRAME] = false;

    optionKeyMap['b'] = OPTION_USE_TEXBO;
    options[OPTION_USE_TEXBO] = true;
    
    optionKeyMap[' '] = OPTION_ANIMATE;
    options[OPTION_ANIMATE] = true;    

    //print the help info
    printf( "Simple_texture_buffer_object - sample showing the usage of GL_EXT_texture_buffer_object\n");
    printf( "  Commands:\n");
    printf( "    q / [ESC] - Quit the application\n");
    printf( "    [SPACE]   - Toggle continuous animation\n");
    printf( "    b         - Toggle using texture buffer objects or regular VBOs\n\n");


    glutDisplayFunc( draw);
    glutIdleFunc( idle);
    glutKeyboardFunc( keys);
    glutSpecialFunc(special);
    glutReshapeFunc( reshape);
    glutMouseFunc( mouse);
    glutMotionFunc( motion);
    glutPassiveMotionFunc(passiveMotion);

    init();

    glutMainLoop();

    return 0;
}
